use itertools::Either;
use std::collections::HashMap;
use std::fmt::{Display, Write};
use std::fs::File;
use std::io::{BufRead, BufReader, Lines};
use std::time::Instant;

// BOILERPLATE
type InputIter = Lines<BufReader<File>>;

fn get_input() -> InputIter {
    let f = File::open("input").unwrap();
    let br = BufReader::new(f);
    br.lines()
}

fn main() {
    let start = Instant::now();
    let ans1 = problem1(get_input());
    let duration = start.elapsed();
    println!("Problem 1 solution: {} [{}s]", ans1, duration.as_secs_f64());

    let start = Instant::now();
    let ans2 = problem2(get_input());
    let duration = start.elapsed();
    println!("Problem 2 solution: {} [{}s]", ans2, duration.as_secs_f64());
}

// PARSE

#[derive(Debug)]
struct Platform {
    rows: Vec<Vec<char>>,
}

impl<T: BufRead> From<Lines<T>> for Platform {
    fn from(lines: Lines<T>) -> Self {
        Self {
            rows: lines.map(|line| line.unwrap().chars().collect()).collect(),
        }
    }
}

enum Direction {
    North,
    South,
    East,
    West,
}

#[derive(Debug, PartialEq)]
enum Axis {
    Rows,
    Columns,
}

impl Direction {
    fn reverse_direction(&self) -> bool {
        match self {
            Direction::North | Direction::West => false,
            Direction::South | Direction::East => true,
        }
    }
    fn axis(&self) -> Axis {
        match self {
            Direction::North | Direction::South => Axis::Columns,
            Direction::East | Direction::West => Axis::Rows,
        }
    }
}

impl<'a> Platform {
    fn width(&self) -> usize {
        self.rows[0].len()
    }
    fn height(&self) -> usize {
        self.col(0).len()
    }
    fn col(&self, col: usize) -> Vec<char> {
        self.rows.iter().map(|row| row[col]).collect()
    }
    fn replace_col(&mut self, idx: usize, col: Vec<char>) {
        for row_idx in 0..col.len() {
            self.rows[row_idx][idx] = col[row_idx];
        }
    }
    fn roll(&mut self, dir: &Direction) {
        match dir.axis() {
            Axis::Columns => {
                for idx in 0..self.width() {
                    self.roll_one(idx, dir)
                }
            }

            Axis::Rows => {
                for idx in 0..self.height() {
                    self.roll_one(idx, dir)
                }
            }
        }
    }
    fn roll_rock_at(old: &mut Vec<char>, inner_idx: usize, dir: &Direction) {
        // Find the first # rock or the edge of the map, we can't look beyond this for a resting position
        match dir.reverse_direction() {
            true => {
                let upper_limit = (inner_idx..old.len())
                    .find(|c| old[*c] == '#')
                    .unwrap_or(old.len());

                if let Some(empty_pos) = (inner_idx..upper_limit).filter(|i| old[*i] == '.').last()
                {
                    old.swap(inner_idx, empty_pos);
                }
            }
            false => {
                // Find the first # rock or the edge of the map, we can't look beyond this for a resting position
                let lower_limit = (0..inner_idx).rev().find(|c| old[*c] == '#').unwrap_or(0);

                if let Some(empty_pos) = (lower_limit..inner_idx)
                    .rev()
                    .filter(|i| old[*i] == '.')
                    .last()
                {
                    old.swap(inner_idx, empty_pos);
                }
            }
        }
    }
    fn roll_one(&mut self, idx: usize, dir: &Direction) {
        let mut old = match dir.axis() {
            Axis::Columns => self.col(idx),
            Axis::Rows => self.rows[idx].clone(),
        };
        let idx_iter: Either<_, _> = match dir.reverse_direction() {
            true => Either::Left((0..old.len()).rev()),
            false => Either::Right(0..old.len()),
        };
        for inner_idx in idx_iter {
            match old[inner_idx] {
                '.' | '#' => continue,
                'O' if dir.reverse_direction() && inner_idx == old.len() => continue,
                'O' if !dir.reverse_direction() && inner_idx == 0 => continue,
                'O' => Self::roll_rock_at(&mut old, inner_idx, dir),
                _ => panic!("invalid character"),
            }
        }
        match dir.axis() {
            Axis::Columns => self.replace_col(idx, old),
            Axis::Rows => self.rows[idx] = old,
        }
    }
    fn score(&self, dir: &Direction) -> u64 {
        match dir.axis() {
            Axis::Columns => self.score_columns(dir),
            Axis::Rows => self.score_rows(dir),
        }
    }
    fn row_or_column_score(&self, idx: usize, dir: &Direction) -> usize {
        if dir.reverse_direction() {
            idx + 1
        } else {
            self.rows.len() - idx
        }
    }
    fn score_columns(&self, dir: &Direction) -> u64 {
        (0..self.height())
            .map(|row| {
                self.rows[row].iter().filter(|c| **c == 'O').count()
                    * self.row_or_column_score(row, dir)
            })
            .sum::<usize>() as u64
    }
    fn roll_cycle(&mut self) {
        self.roll(&Direction::North);
        self.roll(&Direction::West);
        self.roll(&Direction::South);
        self.roll(&Direction::East);
    }

    // find the first loop, return the iteration count when we first saw it and when we saw it again
    fn find_loop(&mut self) -> (usize, usize) {
        let mut first_seen: HashMap<Vec<Vec<char>>, usize> = HashMap::new();
        first_seen.insert(self.rows.clone(), 0);
        let mut i = 0;
        loop {
            self.roll_cycle();
            i += 1;
            if let Some(first_idx) = first_seen.insert(self.rows.clone(), i) {
                return (first_idx, i);
            }
        }
    }
    fn score_rows(&self, dir: &Direction) -> u64 {
        unimplemented!()
    }
}

impl Display for Platform {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        for row in &self.rows {
            for c in row {
                f.write_char(*c)?;
            }
            writeln!(f)?;
        }
        writeln!(f)
    }
}

// PROBLEM 1 solution

fn problem1<T: BufRead>(input: Lines<T>) -> u64 {
    let mut p = Platform::from(input);
    p.roll(&Direction::North);
    p.score(&Direction::North)
}

// PROBLEM 2 solution
fn problem2<T: BufRead>(input: Lines<T>) -> u64 {
    const ITERATIONS: usize = 1000000000;
    let mut p = Platform::from(input);
    let first_loop = p.find_loop();
    let loop_length = first_loop.1 - first_loop.0;
    let cycles_to_skip = ((ITERATIONS - first_loop.0) / loop_length) * loop_length;
    let iterations_remaining = ITERATIONS - first_loop.0 - cycles_to_skip;
    for _ in 0..iterations_remaining {
        p.roll_cycle();
    }

    p.score(&Direction::North)
}

#[cfg(test)]
mod tests {
    use crate::*;
    use std::io::Cursor;

    const EXAMPLE: &str = &"O....#....
O.OO#....#
.....##...
OO.#O....O
.O.....O#.
O.#..O.#.#
..O..#O..O
.......O..
#....###..
#OO..#....";

    #[test]
    fn problem1_example() {
        let c = Cursor::new(EXAMPLE);
        assert_eq!(problem1(c.lines()), 136);
    }

    #[test]
    fn problem2_example() {
        let c = Cursor::new(EXAMPLE);
        assert_eq!(problem2(c.lines()), 64);
    }
}
